{-# LANGUAGE ScopedTypeVariables #-}
{-# LANGUAGE MultiWayIf #-}

-- | Dichotomic search operates by selecting between two distinct alternatives (dichotomies) at each step.
--   Since the range always shrinks but continues to contain the target value until it is found,
--   the method of calculating the division point can vary step by step. This allows us to construct
--   mixed searches such as interleaving interpolation and binary search to get a worst case O(log n)
--   but retain an expected O(log log n) or other situationally appropriate combinations.
module Data.Find.Dichotomic
 ( Divider
 , search, searchM
 , binary, interpolation, interpolationFuzzy, interpolationUnsafe
 ) where

import           Data.Bifunctor
import qualified Data.Find.Linear as Linear
import           Data.Functor.Identity
import           Data.Ratio
import           Data.Word

-- | The function takes the target, the current low side of the range, paired with its value,
--   along wiht the current high side of the range, paired with its value,
--   and returns the next location to divide the range at.
type Divider ix a = a -> (ix, a) -> (ix, a) -> ix

-- | The rule to use binary splitting.
binary :: Integral ix => Divider ix a
binary _ (l, _) (h, _) = max (l+1) $ max (h-1) $ (l+h) `div` 2

-- | The rule to use linear interpolation to estimate the possition of the target,
--   based an assumption of uniform key distribution.
--
--   I believe this needs the ix calculation to be in size of max(size(key), size(value))^2
interpolation :: forall ix s a . (Integral s, Integral ix) => (a -> s) -> Divider ix a
interpolation j t lc hc = fromIntegral $ interpolationUnsafe (toInteger . j) t (first toInteger lc) (first toInteger hc)

-- | The unsafe version of interpolation. The index is u
--
--   I believe this needs the ix calculation to be in size of max(size(key), size(value))^2
interpolationUnsafe :: forall ix a . (Ord ix, Integral ix) => (a -> ix) -> Divider ix a
interpolationUnsafe j' t (l, vl) (h, vh) =
    -- We calculate our estimated position and force it to be inside the existing edges since
    -- our invariant means we've already checked them.
    max (l+1) $ min (h-1) $ floor $ (l%1) + (((j t) - (j vl)) * ((h-l)%1) / ((j vh) - (j vl)))
  where
    j :: a -> Ratio ix
    j = fromIntegral . j'

-- | This variant of interpolation search is optimized for large values in relatively small arrays.
--   It works by finding the first difference in the values, and using that difference to do the interpolation.
--   This works because the high-order difference should be representational of the correct search
--   position and the index, being smaller, doesn't need the full accuracy.
--
--   It peals off Word32s so that we can do most of our math in Word64 without range issues.
--   This is only good if the conversion of the 'a's to Integers is quite expensive.
interpolationFuzzy :: forall ix a . (Ord ix, Integral ix) => (a -> [Word32]) -> Divider ix a
interpolationFuzzy explode t (l, vl) (h, vh)=
  max (l+1) $ min (h-1) $ l + (fromIntegral $ ((fromIntegral $ h-l)*(td - vld)) `div` (vhd-vld))
  where
    -- Get the first part that differs
    (vld, td, vhd) = firstDifference (explode vl) (explode t) (explode vh)
    firstDifference :: [Word32] -> [Word32] -> [Word32] -> (Word64, Word64, Word64)
    -- If it can't be in the range, we can't early terminate but we can maximize the rate
    -- at which we run out of range to search, which we do by splitting it in half.
    firstDifference [] [] [] = (0, 1, 2)
    -- If the target runs out without a difference it can't exist in the range.
    -- (Invariant violation)
    firstDifference _  []  _ = (0, 1, 2)
    -- "reverse" lexicographical case
    firstDifference e0  e1  [] = firstDifference e0 e1 [maxBound]
    -- lexicographical case
    firstDifference []  e1  e2 = firstDifference [minBound] e1 e2
    firstDifference (h0:t0) (h1:t1) (h2:t2) =
      if h0==h1 && h1==h2
      then firstDifference t0 t1 t2
      else (fromIntegral h0, fromIntegral h1, fromIntegral h2)

-- | Conduct a search for a target value using a pure read operation.
search :: forall ix a
        . (Ord ix, Num ix, Ord a)
       => (ix -> a) -- ^ read value at position.
       -> Word -- ^ when <= this many values are left, switch to linear search.
       -> [Divider ix a]
        -- ^ A series of ways to select the search's next division point.
        --   The list will be cycled if its end is reached.
       -> (ix, ix) -- ^ (min, max) of the range to search in.
       -> a -- ^ The value to search for.
       -> Maybe ix
search r ss pgens rng t = runIdentity $ searchM (pure . r) ss pgens rng t

-- | Conduct a search for a target value allowing a monadic read operation.
searchM :: forall m ix a
         . (Monad m, Ord ix, Num ix, Ord a)
        => (ix -> m a) -- ^ read value at position.
        -> Word -- ^ when <= this many values are left, switch to linear search.
        -> [Divider ix a]
        -- ^ A series of ways to select the search's next division point.
        --   The list will be cycled if its end is reached.
        -> (ix, ix) -- ^ (min, max) of the range to search in.
        -> a -- ^ The value to search for.
        -> m (Maybe ix)
searchM r scanSize' pgens (rmin, rmax) t = do
    -- Some dichotomic searches require the value range, so we gather the initial one to pass along.
    vl <- r rmin
    vh <- r rmax
    -- We check our initial invariant that the target isn't one of the edges,
    -- and that it can exist inside the range.
    if | vl == t -> pure $ Just rmin
       | vh == t -> pure $ Just rmax
       | t < vl || vh < t -> pure $ Nothing
       | otherwise -> go (cycle pgens) (rmin, vl) (rmax, vh)
  where
    scanSize = fromIntegral scanSize'
    go :: [Divider ix a] -> (ix, a) -> (ix, a) -> m (Maybe ix)
    go _ (l, _) (h, _) | l+scanSize >= h = Linear.searchM r (l, h) t
    go gens lc hc = do
      let p = (head gens) t lc hc
      cutVal <- r p
      case t `compare` cutVal of
        EQ -> pure $ Just p
        LT -> go (tail gens) lc (p, cutVal)
        GT -> go (tail gens) (p, cutVal) hc
