{-# LANGUAGE LambdaCase #-}
{-# LANGUAGE ScopedTypeVariables #-}
module Main where

import           Control.Monad
import qualified Data.Find.Dichotomic    as Dichotomic
import qualified Data.Find.Interpolation as Interpolation
import qualified Data.Find.Linear        as Linear
import           Data.Functor.Identity
import           Data.List
import           Data.Word
import qualified Data.Vector as V
import           Test.Tasty
import           Test.Tasty.QuickCheck ((===), (==>)) 
import qualified Test.Tasty.QuickCheck as QC

main :: IO ()
main = defaultMain tests

tests :: TestTree
tests =
  testGroup "Find"
    [ linearTests
    , interpolationTests
    , dichotomicTests
    ]

linearTests :: TestTree
linearTests =
  testGroup "linear"
    [ QC.testProperty "search" $ \(valList::[Word64]) ->
        let v = V.fromList $ sort valList
            p = (`map` valList) $ \t ->
                  (Linear.search (v V.!) (0, V.length v-1) t, t)
         in all (\case
                    (Nothing, _) -> False
                    (Just tl, t) -> (v V.! tl) == t) p
   ,  QC.testProperty "searchM" $ \(valList::[Word64]) -> runIdentity $ do
        let v = V.fromList $ sort valList
        p <- forM valList $ \t -> do
               mtl <- Linear.searchM (pure . (v V.!)) (0, V.length v-1) t
               pure (mtl, t)
        pure $ all (\case
                       (Nothing, _) -> False
                       (Just tl, t) -> (v V.! tl) == t) p
   ]

interpolationTests :: TestTree
interpolationTests =
  testGroup "interpolation"
    [ QC.testProperty "rule numeric stability" $ \(t::Integer) (l'::Integer, vl') (h', vh') ->
        let [l, h]   = sort [l', h']
            [vl, vh] = sort [vl', vh']
        in and [vl < t, t < vh, l < h, vl < vh] ==>
        (toInteger $ Dichotomic.interpolation id t (l, vl) (h, vh)) ===
        (Dichotomic.interpolation toInteger (toInteger t) (toInteger l, toInteger vl) (toInteger h, toInteger vh))
    , QC.testProperty "search" $ \(valList::[Word64]) ->
        let v = V.fromList $ sort valList
            p = (`map` valList) $ \t ->
                  (Interpolation.search ((v V.!) . fromIntegral) fromIntegral (0, toInteger $ V.length v-1) t, t)
         in all (\case
                    (Nothing, _) -> False
                    (Just tl, t) -> (v V.! (fromIntegral tl)) == t) p
   ,  QC.testProperty "searchM" $ \(valList::[Word64]) -> runIdentity $ do
        let v = V.fromList $ sort valList
        p <- forM valList $ \t -> do
               mtl <- Interpolation.searchM (pure . (v V.!) . fromIntegral) fromIntegral (0, toInteger $ V.length v-1) t
               pure (mtl, t)
        pure $ all (\case
                       (Nothing, _) -> False
                       (Just tl, t) -> (v V.! (fromIntegral tl)) == t) p
   ]

dichotomicTests :: TestTree
dichotomicTests = testGroup "dichotomic"
 [ testGroup "interpolation"
    [ QC.testProperty "search" $ \(valList::[Word64]) ->
        let v = V.fromList $ sort valList
            p = (`map` valList) $ \t ->
                  (Dichotomic.search (v V.!) 0 [Dichotomic.interpolation id] (0, V.length v-1) t, t)
         in all (\case
                    (Nothing, _) -> False
                    (Just tl, t) -> (v V.! (fromIntegral tl)) == t) p
   ,  QC.testProperty "searchM" $ \(valList::[Word64]) -> runIdentity $ do
        let v = V.fromList $ sort valList
        p <- forM valList $ \t -> do
               mtl <- Dichotomic.searchM (pure . (v V.!)) 0 [Dichotomic.interpolation id] (0, V.length v-1) t
               pure (mtl, t)
        pure $ all (\case
                       (Nothing, _) -> False
                       (Just tl, t) -> (v V.! tl) == t) p
   ]
 , testGroup "interpolationFuzzy"
    [ QC.testProperty "search" $ \(valList::[[Word32]]) ->
        let v = V.fromList $ sort valList
            p = (`map` valList) $ \t ->
                  (Dichotomic.search ((v V.!) . fromIntegral) 0 [Dichotomic.interpolationFuzzy id] (0, toInteger $ V.length v-1) t, t)
         in all (\case
                    (Nothing, _) -> False
                    (Just tl, t) -> (v V.! (fromIntegral tl)) == t) p
   ,  QC.testProperty "searchM" $ \(valList::[[Word32]]) -> runIdentity $ do
        let v = V.fromList $ sort valList
        p <- forM valList $ \t -> do
               mtl <- Dichotomic.searchM (pure . (v V.!) . fromIntegral) 0 [Dichotomic.interpolationFuzzy id] (0, toInteger $ V.length v-1) t
               pure (mtl, t)
        pure $ all (\case
                       (Nothing, _) -> False
                       (Just tl, t) -> (v V.! (fromIntegral tl)) == t) p
   ]
 , testGroup "binary"
    [ QC.testProperty "search" $ \(valList::[Word64]) ->
        let v = V.fromList $ sort valList
            p = (`map` valList) $ \t ->
                  (Dichotomic.search ((v V.!) . fromIntegral) 0 [Dichotomic.binary] (0, toInteger $ V.length v-1) t, t)
         in all (\case
                    (Nothing, _) -> False
                    (Just tl, t) -> (v V.! (fromIntegral tl)) == t) p
   ,  QC.testProperty "searchM" $ \(valList::[Word64]) -> runIdentity $ do
        let v = V.fromList $ sort valList
        p <- forM valList $ \t -> do
               mtl <- Dichotomic.searchM (pure . (v V.!) . fromIntegral) 0 [Dichotomic.binary] (0, toInteger $ V.length v-1) t
               pure (mtl, t)
        pure $ all (\case
                       (Nothing, _) -> False
                       (Just tl, t) -> (v V.! (fromIntegral tl)) == t) p
   ]
 ]
